# FILE: cmdpkg.tcl
#
# Later, come back and make use of cmdpkg in this file.
#
# Also add documentation fields
#
# The command 'em_cmdpkg'
#
# Other notes
# There are consistency rules to be maintained.  Each file should belong
# to only one em_cmdpkg.  Identical (name, version) pairs should reference
# files of *identical* content.  (For example, files in different versions
# of an extension, yet same version of em_cmdpkg)  Identical pfx, ibase, or 
# globalarray values must imply identical name value so that only one can 
# be loaded at a time.
#
# Addition em_cmdpkg data useful for keeping track of:
# author/maintainer/contact
# date
# organization
# usage
# description
# web site
# global litter
# copyright info
# license info
# other documentation, possibly categorized by audience
#
# debugging enable/disable/level
#
# What about multiple commands in a single em_cmdpkg?
#	perhaps possible -- class boilerplate operates on cp
#	just need lookup to return list and then pull command
#	names from it?	Not as elegant, but perhaps possible.

# Only define em_cmdpkg (and support) once (bootstrap protection)
if {![string length [info procs em_cmdpkg]]} {

# Be sure we work on a global array, and define an alias for use within
# this file.
global _em_cmdpkg
upvar #0 _em_cmdpkg cmdvars

# The default values for the keys, parameters, and options to em_cmdpkg
#
# Figure out how to require values
if {![info exists cmdvars]} {
  array set cmdvars {
    @extension		{}
    @file		{}
    #pkgbase		{[file rootname $state(@file)]}
    #cmdbase		{[file rootname $state(@file)]}
    #prefix		{[set tmp [string tolower [file rootname \
			      $state(@file)]]
			  regexp {([a-z])(.*)} $tmp match pfx rest
			  regsub -all {[aeiou0-9_-]} $rest {} rest
			  set second [string range $rest 0 0]
			  regsub -all $second $rest {} rest
			  append pfx $second
			  append pfx [string range $rest 0 0]
			  set pfx ]}
    #version            0.0.0.0.0.0
    -debug		0
    _nextinstance	0
  }
}

# The command defined by this package. V-- the cmdpkg argument
proc em_cmdpkg args "eval _em_cp_Hbase cp \$args"

# A hidden routine which makes things work
# 1. Creation
#       record of (extension, file) pair
#       Set keys: @extension, @file
#
#       Maintain:
#       One cmdpkg per (extension, file) pair.
#
;proc _em_cp_Hbase {cp args} {
  upvar #0 _em_cmdpkg cmdvars
  if {[llength $args]} {
    set firstarg [lindex $args 0]
    if {![string match {[-@]*} $firstarg]} {
# Might want to make this accept unique abbreviations of class methods
# Also might want to make this chunk of code a proc.
      # class methods - only those explicitly defined
      if {[info commands _em_cp_C$firstarg] != {}} {
        return [eval _em_cp_C$firstarg $cp [lrange $args 1 end]]
      } else {
# Add an elseif clause here to probe the cp instance when available
      # class methods - if none, give help
        set cmethods {}
        foreach rawcmd [info commands _em_cp_C*] {
          regexp ^_em_cp_C(.+)$ $rawcmd match cm
          lappend cmethods $cm
        }
        if {[llength $cmethods]} {
          error "em_cmdpkg usage:
	em_cmdpkg ?key_or_option value ...?
Valid keys: [array names cmdvars @*]
Valid options: [array names cmdvars -*]

	em_cmdpkg class_method ?arg ...?
Valid class methods: $cmethods"
        } else {
          error "em_cmdpkg usage:
	em_cmdpkg ?key_or_option value ...?
Valid keys: [array names cmdvars @*]
Valid options: [array names cmdvars -*]"
        }
      }
    }
  }
  # At this point, args is either empty or appears to be list of 
  # option and key values.  Perform basic syntax check, so that array
  # construction won't fail.
  if {[llength $args] % 2} {
    error "em_cmdpkg usage:
	em_cmdpkg ?key_or_option value ...?
Valid keys: [array names cmdvars @*]
Valid options: [array names cmdvars -*]"
  }
  # Make an array out of the arguments
  array set initializers $args
  # Check it for syntactically bogus options/keys
  set bogus {}
  foreach opt [array names initializers] {
    if {[regexp {^[^@-].*} $opt] || ![info exists cmdvars($opt)] } {
      lappend bogus $opt
    }
  }
  if {[llength $bogus]} {
    error "em_cmdpkg error:
	invalid key(s) or option(s): $bogus
Valid keys: [array names cmdvars @*]
Valid options: [array names cmdvars -*]"
  }
  # OK, options and keys are well-formed and valid.
  # 'initializers' array holds them.
  # create set of lookup keys
  array set lookupkeys [array get cmdvars @*]
  array set lookupkeys [array get initializers @*]
  # and evaluate any scripts to determine definite values
  foreach key [array names lookupkeys] {
    if {[regexp {^\[.+\]$} $lookupkeys($key)]} {
      eval set lookupkeys($key) $lookupkeys($key)
    }
    if {![string length $lookupkeys($key)]} {
      error "em_cmdpkg error:
	Missing required key: $key"
    }
  }
  # check for uniqueness
  set this [eval em_cmdpkg lookup [array get lookupkeys]]
  if {[string length $this]} {
    # already have an instance for these keys, reconfigure and return
    eval $this configure [array get initializers -*]
    return $this
  }

  # Create a new instance for these keys
  # Make a global array to hold the instance state.
  set this _em_cmdpkg$cmdvars(_nextinstance)
  incr cmdvars(_nextinstance)
  global $this
  upvar #0 $this state
  if {[info exists $this]} {
    if {[string match _em_cmdpkg3 $this] \
        || [string match _em_cmdpkg1 $this] \
        || [string match _em_cmdpkg4 $this] \
        || [string match _em_cmdpkg5 $this] \
        || [string match _em_cmdpkg6 $this] } {
      # Allow redefinition of the cmdpkgs which are part of the bootstrap 
      # (for now)
    } else {
      error "Programming error in 'em_cmdpkg' command:
Attempt to create two copies of same instance: $this"
    }
  }
  # save class info for later
  array set $this "cmdpkg $cp"
  #
  # store the keys and by the keys
  array set $this [array get lookupkeys]
  set keyvallist {}
  # Must use lsort to guarantee reproducible lookup string
  foreach key [lsort [array names lookupkeys]] {
    lappend keyvallist $lookupkeys($key)
  }
  set keyvals [join $keyvallist ,]
  set cmdvars(keys:$keyvals) $this
  proc $this args "eval _em_cp_Hinstance $this \$args"
  eval $this configure [array get cmdvars -*]
  eval $this configure [array get initializers -*]
  set state(defined) 0

# This is the only part which isn't boiler plate
  set state(selected) 0
  set state(loaded) 0

# finally, return the new instance
  return $this
}

#
# Instance support
#

# A hidden routine that makes things work
;proc _em_cp_Hinstance {this args} {
  if  {[llength $args]} {
    set firstarg [lindex $args 0]
# Might want to make this accept unique abbreviations of instance methods
    # First explicit routines
    if {[info commands _em_cp_I$firstarg] != {}} {
      return [eval _em_cp_I$firstarg $this [lrange $args 1 end]]
    # Then hooks into parameters
    } elseif {![catch [list $this parameters #$firstarg] val]} {
      return $val
    # or hooks into keys
    } elseif {![catch [list $this keys @$firstarg] val]} {
      return $val
    }
    set msg "Invalid instance method: $firstarg"
  } else {
    set msg "No instance method given"
  }
  set imethods {}
  foreach rawcmd [info commands _em_cp_I*] {
    regexp ^_em_cp_I(.+)$ $rawcmd match im
    lappend imethods $im
  }
  array set parameters [$this parameters]
  foreach parameter [array names parameters] {
    regexp ^#(.+)$ $parameter match im
    lappend imethods $im
  }
  array set keys [$this keys]
  foreach key [array names keys] {
    regexp ^@(.+)$ $key match im
    lappend imethods $im
  }
  error "em_cmdpkg instance usage:
	\$instance instance_method ?arg ...?
$msg

Valid instance methods: $imethods"
}

# Destroy this instance -- instance suicide -- undo creation
# 7. blow away object completely
;proc _em_cp_Idelete {this args} {
  upvar #0 $this state
  upvar #0 _em_cmdpkg cmdvars
  # check usage
  if {[llength $args]} {
    error "em_cmdpkg instance delete usage:
	\$instance delete"
  }
  # usage OK.  To delete this instance...
  # 0. Be sure, we undefine it.
  $this undefine
  # Then we can undo the steps of its creation...
  # 1. remove it from the lookup table
  array set keys [$this keys]
  if {![string compare [eval em_cmdpkg lookup [array get keys]] $this]} {
    set keyvallist {}
    # Must use lsort to guarantee reproducible lookup string
    foreach key [lsort [array names keys]] {
      lappend keyvallist $keys($key)
    }
    set keyvals [join $keyvallist ,]
    unset cmdvars(keys:$keyvals)
  } else {
    error "em_cmdpkg instance delete error:
	No entry in the lookup table for $this: $keys"
  }
  # 2. Clear its proc
  rename $this {}
  # 3. Clear its state
  unset state
}

# 1.5 Defintion
#       define and store the parameters: #pkgbase, #cmdbase, #prefix, #version
#
;proc _em_cp_Idefine {this args} {
  upvar #0 $this state
  # Do we want to call this multiple times? Yes.
  if {$state(defined)} {
    return $state(defined)
  }
  upvar #0 _em_cmdpkg cmdvars
  array set validparameters [array get cmdvars #*]
  # check usage
  if {[llength $args] % 2} {
    error "em_cmdpkg instance define usage:
	\$instance define ?parameter value ...?
Valid parameters: [array names validparameters]"
  }
  # First use the default values
  array set inputs [array get validparameters]
  # Then make any replacements from the argument list
  array set inputs $args
  # Might want to make this accept abbreviations of parameters
  foreach parameter [array names inputs] {
    if [info exists validparameters($parameter)] {
      set state($parameter) [subst $inputs($parameter)]
    } else {
      error "em_cmdpkg instance define error:
	Invalid parameter: $parameter
Valid parameters: [array names validparameters]"
    }
  }
  # The version for this command package has been set
  # Collect the timestamp on its file, so we can detect changes
  # and increment the version when necessary
  set pe $state(@extension)
  global $pe
  set state(deftime) [file mtime [file join [set ${pe}(_directory)] \
      $state(@file)]]
  #
  # Enforce the requirements:
  #	* unique pkgbase within extension
  #	* unique cmdbase within extension
  #	* unique prefix within extension
  #	
  # Break into extension for efficiency.  Use ...
  upvar #0 $state(@extension) extinst
  set pep $extinst(_prefix)
  # ...instead of
  # set pep [$state(@extension) cget -prefix]
  set init [string toupper [string range $state(#pkgbase) 0 0]]
  set rest [string tolower [string range $state(#pkgbase) 1 end]]
  set state(name) 		$pep$init$rest
  set state(_command) 		${pep}_$state(#cmdbase)
  # Break into extension for efficiency.  Use ...
  foreach cmdpkg $extinst(_cmdpkgs) {
  # ...instead of
  # foreach cmdpkg [$pe cmdpkgs] {		}
    upvar #0 $cmdpkg other
    if {![string compare $other(#pkgbase) $state(#pkgbase)]} {
      error "em_cmdpkg instance define error:
	Extension '[$pe name]', defined in [$pe directory], 
	includes two conflicting command packages.

        The file '[$cmdpkg file]' defines a command package
	with the name '[$cmdpkg name]'.

	The file '$state(@file)' also defines a command package
	with the name '$state(name)'.

	Within an extension, each command package must have a unique name.

\tEdit the contents file [file join [$pe directory] [$pe contentsFile]]
\tor the addenda file [file join [$pe directory] [$pe addendaFile]]
	 to correct this problem."
    }
    if {![string compare $other(#cmdbase) $state(#cmdbase)]} {
      error "em_cmdpkg instance define error:
	Extension '[$pe name]', defined in [$pe directory], 
	includes two conflicting command packages.

	'[$cmdpkg name]', defined in the file '[$cmdpkg file]', 
	defines the command '[$cmdpkg command]'.

	'$state(name)', defined in the file '[$this file]', 
	also defines the command '$state(_command)'.

	Within an extension, each command package must define a unique
	public command.  

\tEdit the contents file [file join [$pe directory] [$pe contentsFile]]
\tor the addenda file [file join [$pe directory] [$pe addendaFile]]
	 to correct this problem."
    }
    if {![string compare $other(#prefix) $state(#prefix)]} {
      append state(#prefix) $cmdvars(_nextinstance)
      incr cmdvars(_nextinstance)
    }
  }
  # all consistency checks complete.  Define the internal variables
  set state(_globalarray) 	_${pep}_$state(#cmdbase)
  set state(_ibase) 		_${pep}_$state(#cmdbase)
  set state(_hidpfx)		_${pep}_$state(#prefix)_H
  set state(_pcmpfx)		_${pep}_$state(#prefix)_PC
  set state(_pimpfx)		_${pep}_$state(#prefix)_PI
  set state(_cmpfx)		_${pep}_$state(#prefix)_C
  set state(_impfx)		_${pep}_$state(#prefix)_I
  set state(_log)		[em_extension ApplicationLog]
  set state(defined) 1
  $this defined
}  

# 6. undefine
#       clear way for new definition.
;proc _em_cp_Iundefine {this args} {
  upvar #0 $this state
  upvar #0 _em_cmdpkg cmdvars
  if {[llength $args]} {
    error "em_cmdpkg instance undefine usage:
        \$instance undefine"
  }
  if {![$this defined]} {
    return
  }
  $this unselect
  $state(_log) Append "Undefining cmdpkg $this ($state(name))..."
  set state(defined) 0
  set status [$state(@extension) status]
  if {[string match loaded $status] || [string match selected $status]
      || [string match defined $status]} {
    $state(@extension) undefine
  }
  array set validparameters [array get cmdvars #*]
  foreach parameter [array names cmdvars #*] {
     unset state($parameter)
  }
  unset state(deftime)
  unset state(name)
  unset state(_command)
  unset state(_globalarray)
  unset state(_ibase)
  unset state(_hidpfx)
  unset state(_pcmpfx)
  unset state(_pimpfx)
  unset state(_cmpfx)
  unset state(_impfx)
  unset state(_log)
  $this defined
}

# 2. selection
#
#	For cmdpkgs, 'selection' means loading the pacakge database
#	(with the 'package ifneeded' command).
#	Should be only one entry in the package database per
#	cmdpkg package name.
#
#	Should select be callable twice?  No. 
#		(Only if unselect between calls)
#
;proc _em_cp_Iselect {this args} {
  upvar #0 $this state
  # check usage   
  if {[llength $args]} {
    error "em_cmdpkg instance select usage:
        \$instance select"
  }
  if {![$this defined]} {
    error "em_cmdpkg instance select error:
        '\$instance select' called before '\$instance define' or
        after '\$instance undefine'.  Leave it to Tcl's package 
        mecahnism to call '\$instance select' at the right time."
  }
  if {[llength [package versions [$this name]]]} {
    set srcfile [file join [[$this extension] directory] [$this file]]
    error "em_cmdpkg instance select error:
\tCan't select command package '[$this name]', version [$this version],
\tdefined by $srcfile.

	A package by that name already exists in the package database."
  }
  package ifneeded [$this name] [$this version] [list $this load]
  set state(selected) 1
  $this selected
}

# 5. unselect
#
;proc _em_cp_Iunselect {this args} {
  upvar #0 $this state
  # check usage   
  if {[llength $args]} {
    error "em_cmdpkg instance unselect usage:
        \$instance unselect"
  }
  if {![$this selected]} {
    return
  }
  $this unload
  $state(_log) Append "Unselecting cmdpkg $this ($state(name))..."
  set state(selected) 0
  # We only allow one cmdpkg per package name, so no extra
  # forgetting here.
  package forget [$this name]
  $this selected
}

# 3. load
#       source the file. -- defines a command
#       Done at extension level by $ext load command which maintains
#       consistency.
#
;proc _em_cp_Iload {this args} {
  upvar #0 $this state
  # check usage
  if {[llength $args]} {
    error "
'<em_cmdpkg instance> load' usage:
        \$instance load"
  }
  if {!$state(selected)} {
    error "
'<em_cmdpkg instance> load' error:
        '\$instance load' called before '\$instance select' or
        after '\$instance unselect'.  Leave it to Tcl's package
        mecahnism to call '\$instance load' at the right time."
  }
  set pe $state(@extension)
  global $pe
  set srcfile [file join [set ${pe}(_directory)] $state(@file)]

  # want to check on whether we're up to date.
  if {$state(deftime) < [file mtime $srcfile]} {
    error "
'<em_cmdpkg instance> load' error:
	The file '$srcfile'
	has been modified since the extension '[$pe cget -name]' (version [$pe version]) 
	was defined in this application."
  }

  set state(loaded) 1
  # Here it is, the sourcing which loads the command package.
  # $state(_log) Append "Sourcing $srcfile..."
  set myname $state(name)
  set myversion $state(#version)
  if {[catch {uplevel #0 [list source $srcfile]} emsg]} {
    error "
'<em_cmdpkg instance> load' error:
	While loading package $myname, $myversion
	Error sourcing $srcfile :
	$emsg"
  }
  # Inform the package database we've loaded the package
  package provide [$this name] [$this version]
  return $state(loaded)
}

# 4. unload
#       undo sourcing operation. clear commands and arrays.
#
;proc _em_cp_Iunload {this args} {
  upvar #0 $this state
  if {[llength $args]} {
    error "em_cmdpkg instance unload usage:
        \$instance unload"
  }
  if {!$state(loaded)} {
    return
  }
  # Once the (an?) 'em' extension has been loaded, disallow its unloading.
  # Necessary because it is self-dependent, and requires elaborate
  # boot strapping to load.  Redfinition of the 'em' extension require
  # a complete restart of a new Tcl interpreter.
  if {[string match em [$state(@extension) cget -name]]} {
    error "Extension Manager internal error:
	Attempt to unload the extension 'em'"
  }
  $state(_log) Append "Unloading cmdpkg $this ($state(name))..."
  set state(loaded) 0
  if {![string match {} [info commands [$this command]]]} {
    if {![catch {[$this command] instances} instances]} {
      foreach i $instances {
        $i delete
      }
    }
  }
  global [$this globalarray]
  catch {unset [$this globalarray]}
  catch {rename [$this command] {}}
}

;proc _em_cp_Clookup {cp args} {
  upvar #0 _em_cmdpkg cmdvars

  # List all the required keys for this cmdpkg
  set keys [array names cmdvars @*]

  # Check usage
  if {[llength $args] % 2} {
    error "em_cmdpkg lookup usage:
	em_cmdpkg lookup key value ...
Required keys: $keys"
  }
  array set lookupkeys $args
  foreach key $keys {
    if {![info exists lookupkeys($key)]} {
      error "em_cmdpkg lookup error:
	Missing required lookup key: $key
Required keys: $keys
Provided keys: [array names lookupkeys]"
    }
  }

  # For now, this forgives (ignores) extra (key, value) pairs

  # All required keys present.  How many is that?

  if {![llength $keys]} {
    # If none, perform no lookup -- always legal to create new instance
    return {}
  }
  set keyvallist {}
  # Must use lsort to guarantee reproducible lookup string
  foreach key [lsort $keys] {
    lappend keyvallist $lookupkeys($key)
  }
  set keyvals [join $keyvallist ,]
  if {[info exists cmdvars(keys:$keyvals)]} {
    return $cmdvars(keys:$keyvals)
  } else {
    return {}
  }
}

# Perhaps do something more sophisticated here?
#	This depends on adherence to the naming conventions
;proc _em_cp_Cinstances {cp args} {
  info commands _em_cmdpkg*
}

;proc _em_cp_Iclass {this args} {
  return em_cmdpkg
}

;proc _em_cp_Iconfigure {this args} {
  upvar #0 $this state
  upvar #0 _em_cmdpkg cmdvars
  array set validoptions [array get cmdvars -*]
  if {[llength $args] % 2} {
    error "em_cmdpkg instance configure usage:
	\$instance configure ?option value ...?
Valid options: [array names validoptions]"
  }
  array set inputs $args
# Might want to make this accept abbreviations of options
  foreach option [array names inputs] {
    if [info exists validoptions($option)] {
      if {[regexp {^\[.+\]$} $inputs($option)]} {
        # allow initialization by script as well as value
        eval set state($option) $inputs($option)
      } else {
        set state($option) $inputs($option)
      }
    } else {
      error "em_cmdpkg instance configure error:
	Invalid option: $option
Valid options: [array names validoptions]"
    }
  }
}

;proc _em_cp_Icget {this args} {
  upvar #0 $this state
  if {[llength $args] > 1} {
    error "em_cmdpkg instance cget usage:
	\$instance cget ?option?"
  }
  if {[llength $args]} {
    if {[string match -* $args] && [info exists state($args)]} {
      return $state($args)
    } else {
      error "em_cmdpkg instance cget error:
	Invalid option: $args
Valid options: [array names state -*]"
    }
  }
  array get state -*
}

;proc _em_cp_Ikeys {this args} {
  upvar #0 $this state
  if {[llength $args] > 1} {
    error "em_cmdpkg instance keys usage:
	\$instance keys ?option?"
  }
  if {[llength $args]} {
    if {[string match @* $args] && [info exists state($args)]} {
      return $state($args)
    } else {
      error "em_cmdpkg instance keys error:
	Invalid key: $args
Valid keys: [array names state @*]"
    }
  }
  array get state @*
}

;proc _em_cp_Iparameters {this args} {
  upvar #0 $this state
  if {[llength $args] > 1} {
    error "em_cmdpkg instance parameters usage:
	\$instance parameters ?option?"
  }
  if {[llength $args]} {
    if {[string match #* $args] && [info exists state($args)]} {
      return $state($args)
    } else {
      error "em_cmdpkg instance parameters error:
	Invalid parameter: $args
Valid parameters: [array names state #*]"
    }
  }
  array get state #*
}

foreach pv {defined selected name loaded} {
  ;proc _em_cp_I$pv {this args} "
    upvar #0 \$this state
    return \$state($pv)
  "
}

foreach pv {command globalarray ibase 
            hidpfx pcmpfx pimpfx cmpfx impfx} {
  ;proc _em_cp_I$pv {this args} "
    upvar #0 \$this state
    return \$state(_$pv)
  "
}

}
